@@ -1,32 +1,28 @@
 # -*- coding: utf-8 -*-
-"""
-This module provides a set of functions for pyotb
-"""
+"""This module provides a set of functions for pyotb."""
 import inspect
 import os
 import sys
 import textwrap
 import uuid
-import logging
 from collections import Counter
 
 from .core import (App, Input, Operation, logicalOperation, get_nbchannels)
-
-logger = logging.getLogger()
-"""
-Contains several useful functions base on pyotb
-"""
+from .helpers import logger
 
 
 def where(cond, x, y):
-    """
-    Functionally similar to numpy.where. Where cond is True (!=0), returns x. Else returns y
+    """Functionally similar to numpy.where. Where cond is True (!=0), returns x. Else returns y.
+
+    Args:
+        cond: condition, must be a raster (filepath, App, Operation...). If cond is monoband whereas x or y are
+              multiband, cond channels are expanded to match x & y ones.
+        x: value if cond is True. Can be float, int, App, filepath, Operation...
+        y: value if cond is False. Can be float, int, App, filepath, Operation...
+
+    Returns:
+        an output where pixels are x if cond is True, else y
 
-    :param cond: condition, must be a raster (filepath, App, Operation...). If cond is monoband whereas x or y are
-                 multiband, cond channels are expanded to match x & y ones.
-    :param x: value if cond is True. Can be float, int, App, filepath, Operation...
-    :param y: value if cond is False. Can be float, int, App, filepath, Operation...
-    :return: an output where pixels are x if cond is True, else y
     """
     # Checking the number of bands of rasters. Several cases :
     # - if cond is monoband, x and y can be multibands. Then cond will adapt to match x and y nb of bands
@@ -66,13 +62,16 @@ def where(cond, x, y):
 
 
 def clip(a, a_min, a_max):
-    """
-    Clip values of image in a range of values
+    """Clip values of image in a range of values.
+
+    Args:
+        a: input raster, can be filepath or any pyotb object
+        a_min: minimum value of the range
+        a_max: maximum value of the range
+
+    Returns:
+        raster whose values are clipped in the range
 
-    :param a: input raster, can be filepath or any pyotb object
-    :param a_min: minimum value of the range
-    :param a_max: maximum value of the range
-    :return: raster whose values are clipped in the range
     """
     if isinstance(a, str):
         a = Input(a)
@@ -83,16 +82,31 @@ def clip(a, a_min, a_max):
 
 
 def all(*inputs):  # pylint: disable=redefined-builtin
-    """
+    """Check if value is different than 0 everywhere along the band axis.
+
     For only one image, this function checks that all bands of the image are True (i.e. !=0) and outputs
     a singleband boolean raster
     For several images, this function checks that all images are True (i.e. !=0) and outputs
     a boolean raster, with as many bands as the inputs
-    :param inputs:
-    :return: AND intersection
+
+    Args:
+        inputs: inputs can be 1) a single image or 2) several images, either passed as separate arguments
+                or inside a list
+
+    Returns:
+        AND intersection
+
     """
+    # If necessary, flatten inputs
+    if len(inputs) == 1 and isinstance(inputs[0], (list, tuple)):
+        inputs = inputs[0]
+
+    # Add support for generator inputs (to have the same behavior as built-in `all` function)
+    if isinstance(inputs, tuple) and len(inputs) == 1 and inspect.isgenerator(inputs[0]):
+        inputs = list(inputs[0])
+
     # Transforming potential filepaths to pyotb objects
-    inputs = [Input(input) if isinstance(input, str) else input for input in inputs]
+    inputs = [Input(inp) if isinstance(inp, str) else inp for inp in inputs]
 
     # Checking that all bands of the single image are True
     if len(inputs) == 1:
@@ -124,16 +138,30 @@ def all(*inputs):  # pylint: disable=redefined-builtin
 
 
 def any(*inputs):  # pylint: disable=redefined-builtin
-    """
+    """Check if value is different than 0 anywhere along the band axis.
+
     For only one image, this function checks that at least one band of the image is True (i.e. !=0) and outputs
     a single band boolean raster
     For several images, this function checks that at least one of the images is True (i.e. !=0) and outputs
     a boolean raster, with as many bands as the inputs
-    :param inputs:
-    :return: OR intersection
+
+    Args:
+        inputs: inputs can be 1) a single image or 2) several images, either passed as separate arguments
+                or inside a list
+    Returns:
+        OR intersection
+
     """
+    # If necessary, flatten inputs
+    if len(inputs) == 1 and isinstance(inputs[0], (list, tuple)):
+        inputs = inputs[0]
+
+    # Add support for generator inputs (to have the same behavior as built-in `any` function)
+    if isinstance(inputs, tuple) and len(inputs) == 1 and inspect.isgenerator(inputs[0]):
+        inputs = list(inputs[0])
+
     # Transforming potential filepaths to pyotb objects
-    inputs = [Input(input) if isinstance(input, str) else input for input in inputs]
+    inputs = [Input(inp) if isinstance(inp, str) else inp for inp in inputs]
 
     # Checking that at least one band of the image is True
     if len(inputs) == 1:
@@ -165,37 +193,45 @@ def any(*inputs):  # pylint: disable=redefined-builtin
 
 
 def run_tf_function(func):
-    """
-    This function enables using a function that calls some TF operations, with pyotb object as inputs.
+    """This function enables using a function that calls some TF operations, with pyotb object as inputs.
 
     For example, you can write a function that uses TF operations like this :
+        ```python
         @run_tf_function
         def multiply(input1, input2):
             import tensorflow as tf
             return tf.multiply(input1, input2)
 
-    Then you can use it like this :
+        # Then you can use the function like this :
         result = multiply(pyotb_object1, pyotb_object1)  # this is a pyotb object
+        ```
+
+    Args:
+        func: function taking one or several inputs and returning *one* output
+
+    Returns:
+        wrapper: a function that returns a pyotb object
 
-    :param func: function taking one or several inputs and returning *one* output
-    :return wrapper: a function that returns a pyotb object
     """
     try:
         from .apps import TensorflowModelServe
     except ImportError:
-        logger.error('Could not run Tensorflow function: failed to import TensorflowModelServe. Check that you '
-                     'have OTBTF configured (https://github.com/remicres/otbtf#how-to-install)')
+        logger.error('Could not run Tensorflow function: failed to import TensorflowModelServe.'
+                     'Check that you have OTBTF configured (https://github.com/remicres/otbtf#how-to-install)')
         raise
 
     def get_tf_pycmd(output_dir, channels, scalar_inputs):
-        """
-        Create a string containing all python instructions necessary to create and save the Keras model
+        """Create a string containing all python instructions necessary to create and save the Keras model.
 
-        :param output_dir: directory under which to save the model
-        :param channels: list of raster channels (int). Contain `None` entries for non-raster inputs
-        :param scalar_inputs: list of scalars (int/float). Contain `None` entries for non-scalar inputs
-        """
+        Args:
+            output_dir: directory under which to save the model
+            channels: list of raster channels (int). Contain `None` entries for non-raster inputs
+            scalar_inputs: list of scalars (int/float). Contain `None` entries for non-scalar inputs
 
+        Returns:
+            the whole string code for function definition + model saving
+
+        """
         # Getting the string definition of the tf function (e.g. "def multiply(x1, x2):...")
         # TODO: maybe not entirely foolproof, maybe we should use dill instead? but it would add a dependency
         func_def_str = inspect.getsource(func)
@@ -229,14 +265,18 @@ def run_tf_function(func):
         return create_and_save_model_str
 
     def wrapper(*inputs, tmp_dir='/tmp'):
-        """
-        For the user point of view, this function simply applies some TensorFlow operations to some rasters.
+        """For the user point of view, this function simply applies some TensorFlow operations to some rasters.
+
         Implicitly, it saves a .pb model that describe the TF operations, then creates an OTB ModelServe application
         that applies this .pb model to the inputs.
 
-        :param inputs: a list of pyotb objects, filepaths or int/float numbers
-        :param tmp_dir: directory where temporary models can be written
-        :return: a pyotb object, output of TensorFlowModelServe
+        Args:
+            *inputs: a list of pyotb objects, filepaths or int/float numbers
+            tmp_dir: directory where temporary models can be written (Default value = '/tmp')
+
+        Returns:
+            a pyotb object, output of TensorFlowModelServe
+
         """
         # Get infos about the inputs
         channels = []
@@ -269,11 +309,11 @@ def run_tf_function(func):
 
         # Initialize the OTBTF model serving application
         model_serve = TensorflowModelServe({'model.dir': out_savedmodel, 'optim.disabletiling': 'on',
-                                            'model.fullyconv': 'on'}, n_sources=len(raster_inputs), execute=False)
+                                            'model.fullyconv': 'on'}, n_sources=len(raster_inputs), frozen=True)
         # Set parameters and execute
         for i, inp in enumerate(raster_inputs):
             model_serve.set_parameters({f'source{i + 1}.il': [inp]})
-        model_serve.Execute()
+        model_serve.execute()
         # TODO: handle the deletion of the temporary model ?
 
         return model_serve
@@ -283,19 +323,22 @@ def run_tf_function(func):
 
 def define_processing_area(*args, window_rule='intersection', pixel_size_rule='minimal', interpolator='nn',
                            reference_window_input=None, reference_pixel_size_input=None):
-    """
-    Given several inputs, this function handles the potential resampling and cropping to same extent.
+    """Given several inputs, this function handles the potential resampling and cropping to same extent.
+
     WARNING: Not fully implemented / tested
 
-    :param args: list of raster inputs. Can be str (filepath) or pyotb objects
-    :param window_rule: Can be 'intersection', 'union', 'same_as_input', 'specify'
-    :param pixel_size_rule: Can be 'minimal', 'maximal', 'same_as_input', 'specify'
-    :param interpolator: Can be 'bco', 'nn', 'linear'
-    :param reference_window_input: Required if window_rule = 'same_as_input'
-    :param reference_pixel_size_input: Required if pixel_size_rule = 'same_as_input'
-    :return: list of in-memory pyotb objects with all the same resolution, shape and extent
-    """
+    Args:
+        *args: list of raster inputs. Can be str (filepath) or pyotb objects
+        window_rule: Can be 'intersection', 'union', 'same_as_input', 'specify' (Default value = 'intersection')
+        pixel_size_rule: Can be 'minimal', 'maximal', 'same_as_input', 'specify' (Default value = 'minimal')
+        interpolator: Can be 'bco', 'nn', 'linear' (Default value = 'nn')
+        reference_window_input: Required if window_rule = 'same_as_input' (Default value = None)
+        reference_pixel_size_input: Required if pixel_size_rule = 'same_as_input' (Default value = None)
 
+    Returns:
+        list of in-memory pyotb objects with all the same resolution, shape and extent
+
+    """
     # Flatten all args into one list
     inputs = []
     for arg in args:
@@ -364,9 +407,12 @@ def define_processing_area(*args, window_rule='intersection', pixel_size_rule='m
         new_inputs = []
         for inp in inputs:
             try:
-                new_input = App('ExtractROI', {'in': inp, 'mode': 'extent', 'mode.extent.unit': 'phy',
-                                               'mode.extent.ulx': ulx, 'mode.extent.uly': lry,  # bug in OTB <= 7.3 :
-                                               'mode.extent.lrx': lrx, 'mode.extent.lry': uly})  # ULY/LRY are inverted
+                params = {
+                    'in': inp, 'mode': 'extent', 'mode.extent.unit': 'phy',
+                    'mode.extent.ulx': ulx, 'mode.extent.uly': lry,  # bug in OTB <= 7.3 :
+                    'mode.extent.lrx': lrx, 'mode.extent.lry': uly,  # ULY/LRY are inverted
+                }
+                new_input = App('ExtractROI', params)
                 # TODO: OTB 7.4 fixes this bug, how to handle different versions of OTB?
                 new_inputs.append(new_input)
                 # Potentially update the reference inputs for later resampling
@@ -414,7 +460,7 @@ def define_processing_area(*args, window_rule='intersection', pixel_size_rule='m
         inputs = new_inputs
 
         # Update metadatas
-        metadatas = {input: input.GetImageMetaData('out') for input in inputs}
+        metadatas = {inp: inp.GetImageMetaData('out') for inp in inputs}
 
     # Final superimposition to be sure to have the exact same image sizes
     # Getting the sizes of images
@@ -426,14 +472,14 @@ def define_processing_area(*args, window_rule='intersection', pixel_size_rule='m
 
     # Selecting the most frequent image size. It will be used as reference.
     most_common_image_size, _ = Counter(image_sizes.values()).most_common(1)[0]
-    same_size_images = [input for input, image_size in image_sizes.items() if image_size == most_common_image_size]
+    same_size_images = [inp for inp, image_size in image_sizes.items() if image_size == most_common_image_size]
 
     # Superimposition for images that do not have the same size as the others
     new_inputs = []
     for inp in inputs:
         if image_sizes[inp] != most_common_image_size:
-            new_input = App('Superimpose', inr=same_size_images[0], inm=inp, interpolator=interpolator)
-            new_inputs.append(new_input)
+            superimposed = App('Superimpose', inr=same_size_images[0], inm=inp, interpolator=interpolator)
+            new_inputs.append(superimposed)
         else:
             new_inputs.append(inp)
     inputs = new_inputs
